%{
/* 
 * urldecode decodes specified url-encoded or multipart-encoded variables 
 * supplied on stdin, and writes their values to files in a directory, or 
 * (for url-encoded data only) to environment variables.
 *
 * name=value is printed out on stdout if we're parsing url-encoded stuff.
 * If we're parsing multipart-encoded stuff, we just print out name, and
 * filename and/or content-type if it exists. 
 *
 * $Id: urldecode.lex,v 1.19 2010/08/06 20:09:25 jdd Exp $
 */

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <string.h>
#include "patchlevel.h"

/* Input buffer sizes.  BUFFSIZE should be greater than LINESIZE */
#ifndef BUFFSIZE
#define BUFFSIZE 4096
#endif /* BUFFSIZE */
#ifndef LINESIZE
#define LINESIZE 512
#endif /* LINESIZE */

/* default permissions for directories created */
#ifndef FILEMODE
#define FILEMODE 0700
#endif /* FILEMODE */

/* if an environment variable exists of the same name, overwrite it */
#ifndef ENV_OVERWRITE
#define ENV_OVERWRITE 1
#endif /* ENV_OVERWRITE */

/* MIME text */
#define CONTENTDISP "Content-Disposition: form-data; name=\042"
#define CONTENTTYPE "Content-Type: "

char *progname, *t, *f, **vars, *cmd=NULL, *dir=NULL;
FILE *fp;
int numvars, match=0, ismultipart=0;
u_long globalmax=0, bytes=0;

enum mode {
	name,
	body,
	error,
} parsing=name;

/* 
 * strmatch is strcmp if we're not using regular expressions, otherwise
 * it's a sort of regex extension of strcmp where the regular expression
 * is the first argument and the string being matched is the second).
 */
#ifdef REGEX
#define USAGE "Usage: %s [-V | -c command | -d directory | -D directory [ -M maxbytes ]] varname-regex [varname-regex ...]\n"
#include <sys/types.h>
#include <regex.h>
int strmatch(char *expression, char *str){
	regex_t re;
	regmatch_t pm;
	(void)regcomp(&re, expression, REG_EXTENDED|REG_NEWLINE);
	/* 
	 * We need to check to see that the entire string has been
	 * matched to the expression, not just part of it, hence we need
	 * to check the substring matching structure pm, to check
	 * the start and the end of the match, it's not enough to
	 * check the return of regexec, because it will indicate
	 * a match if expression matches a substring of str.
	 */
	if(REG_NOMATCH==regexec(&re, str, (size_t)1, &pm, 0)){
		return(REG_NOMATCH); /* no match at all */
	} else if (0==pm.rm_so && strlen(str)==pm.rm_eo) {
		return(0);  /* matches the whole string */
	} else {
		return(REG_NOMATCH); /* matches a substring only */
	}
} 
#else /* !REGEX */
#define USAGE "Usage: %s [-V | -c command | -d directory | -D directory [ -M maxbytes ]] varname [varname ...]\n"
int strmatch(char *expression, char *str){
	return(strcmp(expression, str));
}
#endif /* !REGEX */

/*
 * Mgetc() - fgetc with checking for max bytes.
 */
int Mgetc(FILE *sm)
{
	int c;
	if(EOF!=(c=fgetc(sm))) bytes++;
	if(globalmax && bytes>globalmax){
		errno=ERANGE;
		return(EOF);
	} else return(c);
}

/*
 * Mgets() - fgets with checking for max bytes.
 */
char *Mgets(char *sg, int sz, FILE *sm)
{
	if(NULL!=fgets(sg,sz,sm)){
		bytes+=strlen(sg);
	}
	if(globalmax && bytes>globalmax){
		errno=ERANGE;
		return((char *)NULL);
	}
	return(sg);
}

/* cat() concatenates the string pt2 onto the end of the string pt1, returning 
 *	(possibly a new) pt1, (re|m)alloced as necessary.  pt1 is assumed to 
 *	be reallocable. 
 */
char *cat(char *pt1, char *pt2){
	/* concatenate pt2 onto the end of pt1, returning (possibly a new) pt1,
	 * (re|m)alloc as necessary. pt1 is assumed to be reallocable. */
	char *p1, *p2;

	if(NULL==(pt1=realloc((void *)pt1, strlen(pt1)+strlen(pt2)+1))){
		errno=ENOSPC;
		perror(progname);
		exit(1);
	}
	(void)strcat(pt1, pt2); 

	return(pt1);
}

/* 
 * Error() takes one string output and puts it out as a stderr message, then
 * exit(1)'s.
 */
void Error(char *msg){
	fprintf(stderr, "%s: %s\n", progname, msg);
	exit(1);
}

void processname(){
	/* process the "name=" part of a urlencoded variable */
	/* variable name */
	if(parsing==name){
		int i;
		for(i=0;i<numvars;i++){
			if(0==strmatch(vars[i],t)){
				printf("%s=", t);
				if(dir){
					f=cat(f,dir);
					f=cat(f,"/");
					f=cat(f,t);
					if(NULL==(fp=fopen(f,"w"))){
						perror(f);
						exit(1);
					}
				} else if(cmd){
					f=cat(f,t);
				}
				match++;
				break;
			}
		}	
		parsing=body;
	}
	t[0]=(char)0; 
}

void processvalue(){
	/* process the "=value" part of a urlencoded variable */
	if(parsing==body && match) {
		printf("%s\n", t); 
		if(dir){
			if(0>fputs(t, fp)){
				perror(f);
				exit(1);
			} 
			if(0>fclose(fp)){
				perror(f);
				exit(1);
			}
			f[0]=(char)0;
		} else if(cmd){
#ifdef USE_PUTENV
			char *e;
			e=cat(f,"=");
			e=cat(e,t);
			if(0>putenv(e)){
				perror(f);
				exit(1);
			}
#else /* !USE_PUTENV */
			setenv(f,t,ENV_OVERWRITE);
#endif /* !USE_PUTENV */
			f[0]=(char)0;
		}
	} 
	t[0]=(char)0;
	parsing=name;
	match=0;
}

void trimnl(char *b){
	/* trim (in place) any \r\n or \n from the end of 
	 * the specified string by replacing with \0
	 */
	if('\n'==b[strlen(b)-1]) b[strlen(b)-1]='\0';
	if('\r'==b[strlen(b)-1]) b[strlen(b)-1]='\0';
}


void multipart(){
	/* Process multipart-encoded data, not url-encoded; completely
	 * different parsing rules.  multipart-encoded looks like this:
	 *
	 * -----------------------------23281168279961
	 * Content-Disposition: form-data; name="name"
	 *
	 * defaulttext
	 * -----------------------------23281168279961
	 * Content-Disposition: form-data; name="user"
	 *
	 * defaulttext
	 * -----------------------------23281168279961
	 * Content-Disposition: form-data; name="myfile"; filename="disks1.ps"
	 * Content-Type: application/postscript
	 * %!PS-Adobe-3.0
	 * ...<possible binary contents here>...
	 * -----------------------------23281168279961
	 * Content-Disposition: form-data; name="thoughts"
	 * 
	 * defaultthoughts
	 * -----------------------------23281168279961--
	 *
	 */
	char buf[BUFFSIZE], headbuf[LINESIZE+2], *header, name[BUFFSIZE];
	int firsttime=1; /* first time through loop? */

	/* the header is the part that looks like this:
	 * "-----------------------------23281168279961".  We distinguish
	 * "header" from headbuf, which contains the header with a 
	 * preceeding cr/ln.  We need "headbuf" because when we're reading
	 * the otherwise binary contents of a file, we won't know that we've
	 * come to the end of the file until we see something like
	 * "\r\n-----------------------------23281168279961".
	 */
	headbuf[0]='\r'; headbuf[1]='\n'; headbuf[2]='\0'; 
	header=headbuf+2;
	if(NULL==Mgets(header,LINESIZE,stdin)) {
		if(ERANGE==errno) {
			Error("Maximum input size exceeded."); 
		} else {
			Error("Header expected, no header present.");
		}
	}
	trimnl(header); 
	while(1){
		int i, hdr, b;
		char *p, *h;
		/* appendix is for stuff after name in content-disposition,
		 * usually only for files.  We print it out after
	         */
		char *appendix; 

		if(firsttime) {
			firsttime=0;
		} else {
			/* next read is either "--" if final EOF, or
			 * a blank line if there is more to read 
			 */
			if(NULL==Mgets(buf, sizeof(buf), stdin)) {
				Error("Premature EOF.");
			}
			if(0==strcmp(buf, "--\r\n")){
				/* we should be all done.  Try to read
				 * some more and if we get anything, throw
				 * an error.
				 */
				if(NULL!=Mgets(buf,sizeof(buf),stdin)){
					Error("Unexpected trailing data.");
				}
				exit(0);
			} else if(0!=strcmp(buf,"\r\n")) {
				Error("Unexpected appendage to header.");
			}
		}

		/* read content-disposition. */
		if(NULL==Mgets(buf, sizeof(buf), stdin)) {
			if(ERANGE==errno) {
				Error("Maximum input size exceeded."); 
			} else {
				Error("Content-Disposition expected.");
			}
		}
		trimnl(buf);
		t=buf+strlen(CONTENTDISP);
		if(0!=strncmp(buf,CONTENTDISP, strlen(CONTENTDISP))){
			Error("Malformed Content-disposition.");
		}
		if(NULL==(appendix=strchr(t, '\042'))) {
			Error("Missing trailing quotation mark.");
		}
		for(i=0;i<numvars;i++){
			if('\042'==(*appendix)) {
				*appendix='\0';
				appendix++;
			}
			if(0==strmatch(vars[i],t)){
				match++;
				if(';'==(*appendix)) appendix++;
				if(' '==(*appendix)) appendix++;
				if(appendix && *appendix){
					printf("%s %s", t, appendix);
				} else {
					trimnl(t);
					printf("%s", t);
				}
				if(dir){
					f=cat(f,dir);
					f=cat(f,"/");
					f=cat(f,t);
					if(NULL==(fp=fopen(f,"w"))){
						perror(f);
						exit(1);
					} 
					free(f); 
					f=malloc(sizeof(char)); 
					*f='\0';
				}
				break;
			}
		}

		/* Read Content-Type if any, and blank line */
		if(NULL==Mgets(buf, sizeof(buf), stdin)){
			if(ERANGE==errno) {
				Error("Maximum input size exceeded."); 
			} else {
				Error("Premature data end after content-disposition.");
			}
		}
		if(0==strncmp(buf, CONTENTTYPE, strlen(CONTENTTYPE))){
			if(match) {
				trimnl(buf+strlen(CONTENTTYPE));
				printf(" contenttype=%s\n", 
					buf+strlen(CONTENTTYPE));
			}
			if(NULL==Mgets(buf, sizeof(buf), stdin)) {
				if(ERANGE==errno) {
					Error("Maximum input size exceeded."); 
				} else {
					Error("Premature data end after content-type.");
				}
			}
		} else if(match) printf("\n");

		/* we should have a blank line now */
		if(0!=strcmp(buf, "\r\n")){
			Error("Unexpected data.");
		}

		/* now comes value -- could be multiple lines. */
		hdr=b=0;
		h=headbuf;
		p=buf;
		while(1){
			int c;
			/*
			 * Compare input stream to headbuf, saving
			 * it as we go.  If we get to the end of
			 * the header and we're still matching, we
			 * have a header. If the match fails before
			 * then, we need to write out the input 
			 * stream and keep going.  p is the pointer
			 * into the input stream, h is the pointer
			 * into the header.
			 */

			/* exit when we've matched the whole header */
			if(hdr>=strlen(headbuf)) break;

			/* sanity check */
			if(p-buf>=sizeof(buf)) Error("Buffer overflow.");

			/* get and save a character */
			if(EOF==(c=Mgetc(stdin))) {
				if(ERANGE==errno) {
					Error("Maximum input size exceeded."); 
				} else Error("Unexpected EOF.");
			}
			*p=(char)c;

			/* primitive progress meter */
			/* if(!(b%1000000)) fprintf(stderr, "."); */
			if(hdr && *p!=*h) {
				/* oops this wasn't a header after all */
				if(match) {
					/* write out what we've saved so far */
					if(0==fwrite(buf,sizeof(char),hdr,fp)){
						Error(strerror(errno));
					}
					b+=hdr; /* byte count */
				}
				hdr=0; h=headbuf; buf[0]=*p; p=buf;
			}
			if(*p==*h){
				/* save this character, might be header */
				p++; h++; hdr++;
			} else if(!hdr) {
				if(match) {
					if(EOF==fputc(*p,fp)){
						/* write out this character */
						Error(strerror(errno));
					}
				}
				p=buf;
			} 
		}
		if(match)(void)fclose(fp);
		match=0;
	}
}

int yywrap(){
	processvalue();
	if(cmd) system(cmd);
	/* yyterminate(); */
	return(1);
}

%}

BASE 	[0-9a-zA-Z_]
EXTRA 	[\.\-\*]
HEX	[0-9A-F]


%%

{BASE}+	{ 
	/* characters in use for both names and values */
	t=cat(t, yytext); 
}

{EXTRA}+ { 
	/* characters only in use for values */
	if(parsing==name) {
		parsing=error;
	} else {
		t=cat(t,yytext);
	}
}

"+" { 
	/* spaces are encoded as plus sign */
	t=cat(t, " ");
}

"=" { 
	/* when we hit an = sign, we are at the end of a var name */
	processname();
}

"&" {
	/* when we hit &, we are at the end of a variable value */
	processvalue();
}


"%0D%0A" {
	/* map DOS and Web-style CRLF line endings to UNIX-style newlines */
	if(parsing==name) {
		parsing=error;
	} else {
		t=cat(t, "\n"); 
	}
}

"%"{HEX}{HEX} { 
	/* urlencoded hex values are e.g. %0F for ascii 15 */ 
	if(parsing==name) {
		parsing=error;
	} else {
		char c[2]; 
		c[0]=(char)strtol((char *)(yytext+1), NULL, 16); 
		c[1]='\0'; 
#ifdef notdef
	/* escape single quotes */
		if(c[0]==(char)(39)){ /* single quote */
			t=cat(t, "\\'");
		} else 
#endif /* notdef */
		t=cat(t, (char *)c);
	
	}
}

[% \n]+  /* ignore newlines and % */ {}

. { 
	/* unknown character */
	parsing=error;
}
%%

int main(int argc, char *argv[]){
	int a;
	progname=argv[0];

	opterr=0;
	while(0<=(a=getopt(argc,argv,"Vc:d:D:M:"))){
		switch(a){
		case 'V':
			/* version information */
			printf("%s\n",patchlevel);
			return(0);
			break;
		case 'c':
			/* run specified command with urlencoded parameters 
			 * as environment variables in its environment.
			 */
			cmd=optarg;
			break;
		case 'D':
			/* Take multipart-encoded input instead of url-encoded;
			 * otherwise is the same as d 
			 */
			ismultipart++;
			/* fall thru here */
		case 'd':
			/* create a new directory (must not already exist)
			 * and put each urlencoded variable as a file in that
			 * directory, the name of which is the variable name, 
			 * and the contents of which are the variable value.
			 */
			dir=optarg;
			if(0!=mkdir(dir, (mode_t)FILEMODE)){
				perror(progname);
				return(1);
			}
			break;
		case 'M':
			/* maximum size of input */
			globalmax=atol(optarg);
			break;
		case '?':
			fprintf(stderr, "%s: Unknown parameter %c\n", 
				progname, optopt);
			/* fall thru here */
		default:
			fprintf(stderr, USAGE, progname);
			return(1);
			break;
		}
	}
	/* list of variables we are interested in */
	numvars=argc-optind;  /* ok if numvars=0, just checking input */
	vars=&(argv[optind]);	
	if(NULL==(t=malloc(sizeof(char)))||NULL==(f=malloc(sizeof(char))))
		Error("Out of memory.");
	t[0]=f[0]=(char)0;
	if(ismultipart) multipart(); else yylex();
}
